--- @classmod Blob
--- @author megagrump
--- @license MIT
--- @copyright 2017 megagrump
local ffi = require('ffi')

local Blob = {}

local band, bor, bnot, shl, shr = bit.band, bit.bor, bit.bnot, bit.lshift, bit.rshift
local _native, _endian
local _tags, _taggedReaders, _taggedWriters, _lengthBytes

--- Initializes a new Blob instance.
-- @param data (nil) A Lua string with which to populate the blob
-- @param size (1024) The initial size of the blob
-- @param options (nil) Blob options
-- Available options:
-- source_endianess: the byte order in the source data. Can be either 'le' or 'be', or nil to use the host's byte
-- order (default)
-- target_endianess: the byte order written data will be saved in. Can be either 'le' or 'be', or nil to use the host's
-- byte order (default)
function Blob:init(data, size, options)
	options = options or {}
	local sourceEndian = options.source_endianess or (ffi.abi('le') and 'le' or 'be')
	local targetEndian = options.target_endianess or (ffi.abi('le') and 'le' or 'be')

	assert(sourceEndian == 'le' or sourceEndian == 'be', "Invalid source endianess")
	assert(targetEndian == 'le' or targetEndian == 'be', "Invalid target endianess")

	self._endian_in = _endian[sourceEndian]._in
	self._endian_out = _endian[targetEndian]._out
	self._length = 0

	size = size or 1024
	size = data and math.max(size, #data) or size
	self:_allocate(size)
	if data then
		ffi.copy(self._data, data, #data)
		self._length = #data
	end
	self._readPtr = 0
end

--- Reads a string, a number, a boolean or a table from the input buffer.
-- The data must have been written with Blob:write().
-- The type of the value is automatically detected from the input metadata.
-- @return The value read
function Blob:read()
	local tag, value = self:_readTagged()
	return value
end

--- Reads a Lua number from the input buffer.
-- @return The number read
function Blob:readNumber()
	_native.u32[0], _native.u32[1] = self:readU32(), self:readU32()
	return _native.n
end

--- Reads a string from the input buffer.
-- The string must have been written with Blob:write() or Blob:writeString().
-- @return The string read
function Blob:readString()
	local len, ptr = self:_readLength(), self._readPtr
	assert(ptr + len - 1 < self._length, "Out of data")
	self._readPtr = ptr + len
	return ffi.string(ffi.cast('uint8_t*', self._data + ptr), len)
end

--- Reads a boolean value from the input buffer.
-- The data is expected to be 8 bits long, 0 == false
-- @return The boolean value read
function Blob:readBool()
	return self:readU8() ~= 0
end

--- Reads a Lua table from the input buffer.
-- The table must have been written with Blob:write() or Blob:writeTable().
-- @return The value read
function Blob:readTable()
	local result = {}
	local tag, key = self:_readTagged()
	while tag ~= _tags.stop do
		tag, result[key] = self:_readTagged()
		tag, key = self:_readTagged()
	end
	return result
end

--- Reads one unsigned 8 bit value from the input buffer.
-- @return The unsigned 8 bit value read
function Blob:readU8()
	assert(self._readPtr < self._length, "Out of data")
	local u8 = self._data[self._readPtr]
	self._readPtr = self._readPtr + 1
	return u8
end

--- Reads one signed 8 bit value from the input buffer.
-- @return The signed 8 bit value read
function Blob:readS8()
	local u8 = self:readU8()
	return -band(u8, 2 ^ 7 - 1) + band(u8, 2 ^ 7)
end

--- Reads one unsigned 16 bit value from the input buffer.
-- @return The unsigned 16 bit value read
function Blob:readU16()
	local ptr = self._readPtr
	assert(ptr + 1 < self._length, "Out of data")
	self._readPtr = ptr + 2
	return self._endian_in._16(self._data[ptr], self._data[ptr + 1])
end

--- Reads one signed 16 bit value from the input buffer.
-- @return The signed 16 bit value read
function Blob:readS16()
	local u16 = self:readU16()
	return -band(u16, 2 ^ 15 - 1) + band(u16, 2 ^ 15)
end

--- Reads one unsigned 32 bit value from the input buffer.
-- @return The unsigned 32 bit value read
function Blob:readU32()
	local ptr = self._readPtr
	assert(ptr + 3 < self._length, "Out of data")
	self._readPtr = ptr + 4
	return self._endian_in._32(self._data[ptr], self._data[ptr + 1], self._data[ptr + 2], self._data[ptr + 3])
end

--- Reads one signed 32 bit value from the input buffer.
-- @return The signed 32 bit value read
function Blob:readS32()
	local u32 = self:readU32()
	return -band(u32, 2 ^ 31 - 1) + band(u32, 2 ^ 31)
end

--- Reads one unsigned 64 bit value from the input buffer.
-- @return The unsigned 64 bit value read
function Blob:readU64()
	local ptr = self._readPtr
	assert(ptr + 7 < self._length, "Out of data")
	self._readPtr = ptr + 8
	return self._endian_in._64(self._data[ptr], self._data[ptr + 1], self._data[ptr + 2], self._data[ptr + 3],
		self._data[ptr + 4], self._data[ptr + 5], self._data[ptr + 6], self._data[ptr + 7])
end

--- Reads one signed 64 bit value from the input buffer.
-- @return The signed 64 bit value read
function Blob:readS64()
	_native.s64 = self:readU64()
	return _native.s64
end

--- Reads one 32 bit floating point value from the input buffer.
-- @return The 32 bit floating point value read
function Blob:readFloat()
	_native.u32[0] = self:readU32()
	return _native.f[0]
end

--- Reads raw binary data from the input buffer.
-- @param len The length of the data (in bytes) to read
-- @return A string with raw data
function Blob:readRaw(len)
	local ptr = self._readPtr
	assert(ptr + len - 1 < self._length, "Out of data")
	self._readPtr = ptr + len
	return ffi.string(ffi.cast('uint8_t*', self._data + ptr), len)
end

-----------------------------------------------------------------------------

--- Writes a value to the output buffer.
-- Supported value types are number, string, bool and table.
-- @param value The value to write
-- @return A reference to the current blob
function Blob:write(value)
	return self:_writeTagged(value)
end

--- Writes a Lua number to the output buffer.
-- @param value The value to write
-- @return A reference to the current blob
function Blob:writeNumber(value)
	_native.n = value
	self:writeU32(_native.u32[0]):writeU32(_native.u32[1])
	return self
end

--- Writes a boolean value to the output buffer.
-- The value is written as an 8 bit value.
-- @param bool The value to write
-- @return A reference to the current blob
function Blob:writeBool(bool)
	self:writeU8(bool and 1 or 0)
	return self
end

--- Writes a string to the output buffer.
-- Stores the length of the string in a compact format before the actual string data.
-- @param str The string to write
-- @return A reference to the current blob
function Blob:writeString(str)
	local length = #str
	local makeRoom = (self._size - self._length) - (length + _lengthBytes(length))
	if makeRoom < 0 then
		self:_grow(math.abs(makeRoom))
	end
	self:_writeLength(length)
	ffi.copy(ffi.cast('char*', self._data + self._length), str, length)
	self._length = self._length + length
	return self
end

--- Writes an unsigned 8 bit value to the output buffer.
-- @param u8 The value to write
-- @return A reference to the current blob
function Blob:writeU8(u8)
	if self._length == self._size then self:_grow() end
	self._data[self._length] = u8
	self._length = self._length + 1
	return self
end

--- Writes a signed 8 bit value to the output buffer.
-- @param s8 The value to write
-- @return A reference to the current blob
function Blob:writeS8(s8)
	return self:writeU8(band(bnot(s8) + 1, 2 ^ 8 - 1))
end

--- Writes an unsigned 16 bit value to the output buffer.
-- @param u16 The value to write
-- @return A reference to the current blob
function Blob:writeU16(u16)
	local len = self._length
	if len + 1 > self._size then self:_grow() end
	local b1, b2 = self._endian_out._16(band(u16, 2 ^ 8 - 1), shr(u16, 8))
	self._data[len], self._data[len + 1] = b1, b2
	self._length = len + 2
	return self
end

--- Writes a signed 16 bit value to the output buffer.
-- @param s16 The value to write
-- @return A reference to the current blob
function Blob:writeS16(s16)
	return self:writeU16(band(bnot(s16) + 1, 2 ^ 16 - 1))
end

--- Writes an unsigned 32 bit value to the output buffer.
-- @param u32 The value to write
-- @return A reference to the current blob
function Blob:writeU32(u32)
	local len = self._length
	if len + 3 > self._size then self:_grow() end
	local w1, w2 = self._endian_out._32(band(u32, 2 ^ 16 - 1), shr(u32, 16))
	local b1, b2 = self._endian_out._16(band(w1, 2 ^ 8 - 1), shr(w1, 8))
	local b3, b4 = self._endian_out._16(band(w2, 2 ^ 8 - 1), shr(w2, 8))
	self._data[len], self._data[len + 1], self._data[len + 2], self._data[len + 3] = b1, b2, b3, b4
	self._length = len + 4
	return self
end

--- Writes a signed 32 bit value to the output buffer.
-- @param s32 The value to write
-- @return A reference to the current blob
function Blob:writeS32(s32)
	return self:writeU32(band(bnot(s32) + 1, 2 ^ 32 - 1))
end

--- Writes an unsigned 64 bit value to the output buffer.
-- @param u64 The value to write
-- @return A reference to the current blob
function Blob:writeU64(u64)
	_native.u64 = u64
	local a, b = self._endian_out._64(_native.u32[0], _native.u32[1])
	return self:writeU32(a):writeU32(b)
end

--- Writes a signed 64 bit value to the output buffer.
-- @param s64 The value to write
-- @return A reference to the current blob
function Blob:writeS64(s64)
	_native.s64 = s64
	local a, b = self._endian_out._64(_native.u32[0], _native.u32[1])
	return self:writeU32(a):writeU32(b)
end

--- Writes a 32 bit floating point value to the output buffer.
-- @param float The value to write
-- @return A reference to the current blob
function Blob:writeFloat(float)
	_native.f[0] = float
	return self:writeU32(_native.u32[0])
end

--- Writes raw data to the output buffer.
-- @param str The data to write
-- @return A reference to the current blob
function Blob:writeRaw(str)
	local length = #str
	local makeRoom = (self._size - self._length) - length
	if makeRoom < 0 then
		self:_grow(math.abs(makeRoom))
	end
	ffi.copy(ffi.cast('char*', self._data + self._length), str, length)
	self._length = self._length + length
	return self
end

--- Writes a table to the output buffer.
-- Supported field types are number, string, bool and table. Cyclic references throw an error.
-- @param t The table to write
-- @return A reference to the current blob
function Blob:writeTable(t)
	return self:_writeTable(t, {})
end

-----------------------------------------------------------------------------

--- Returns the current buffer contents as a string.
-- @return Current buffer contents
function Blob:string()
	return ffi.string(self._data, self._length)
end

--- Returns the number of bytes in the output buffer.
-- @return Number of bytes in the output buffer
function Blob:length()
	return self._length
end

--- Resets the read position to the beginning of the buffer.
-- @return A reference to the current blob
function Blob:rewind()
	self._readPtr = 0
	return self
end

--- Returns the current read position as an offset from the start of the buffer in bytes.
-- @return Read position
function Blob:position()
	return self._readPtr
end

-----------------------------------------------------------------------------

function Blob:_allocate(size)
	local data
	if size > 0 then
		data = ffi.new('uint8_t[?]', size)
		if self._data then
			ffi.copy(data, self._data, self._length)
		end
	end
	self._data, self._size = data, size
	self._length = math.min(size, self._length)
end

function Blob:_grow(minimum)
	minimum = minimum or 0
	local newSize = math.max(self._size + minimum, math.floor(math.max(1, self._size * 1.5) + .5))
	self:_allocate(newSize)
end

function Blob:_readLength()
	local result = self:readU8()
	if band(result, 0x00000080) == 0 then return result end
	result = band(result, 0x0000007f) + self:readU8() * 2 ^ 7
	if band(result, 0x00004000) == 0 then return result end
	result = band(result, 0x00003fff) + self:readU8() * 2 ^ 14
	if band(result, 0x00200000) == 0 then return result end
	result = band(result, 0x001fffff) + self:readU8() * 2 ^ 21
	if band(result, 0x10000000) == 0 then return result end
	return band(result, 0x0fffffff) + self:readU8() * 2 ^ 28
end

function Blob:_readTagged()
	local tag = self:readU8()
	return tag, tag ~= _tags.stop and _taggedReaders[tag](self)
end

function Blob:_writeLength(value)
	for i = 7, 28, 7 do
		local mask, shift = 2 ^ i - 1, i - 7
		if value < 2 ^ i then
			return self:writeU8(shr(band(value, mask), shift))
		else
			self:writeU8(shr(band(value, mask), shift) + 0x80)
		end
	end
	return self:writeU8(shr(band(value, 0xf0000000), 28))
end

function Blob:_writeTable(t, stack)
	assert(type(t) == 'table', string.format("Invalid type '%s' for Blob:writeTable", type(t)))
	assert(not stack[t], string.format("Cycle detected; can't serialize table"))
	stack[t] = t
	for key, value in pairs(t) do
		self:_writeTagged(key, stack)
		self:_writeTagged(value, stack)
	end
	stack[t] = nil
	self:writeU8(_tags.stop)
	return self
end

function Blob:_writeTagged(value, stack)
	local t = type(value)
	local tag = _tags[t]
	assert(tag, string.format("Can't write values of type '%s'", t))
	self:writeU8(tag)
	_taggedWriters[tag](self, value, stack)
	return self
end

function _lengthBytes(length)
	for i = 1, 4 do
		if length < 2 ^ (i * 7) then return i end
	end
	return 5
end

_native = ffi.new[[
	union {
		  int8_t s8[8];
		 uint8_t u8[8];
		 int16_t s16[4];
		uint16_t u16[4];
		 int32_t s32[2];
		uint32_t u32[2];
		   float f[2];
		 int64_t s64;
		uint64_t u64;
		  double n;
	}
]]

_endian = {
	le = {
		_out = {
			_16 = function(b1, b2) return b1, b2 end,
			_32 = function(w1, w2) return w1, w2 end,
			_64 = function(d1, d2) return d1, d2 end,
		},

		_in = {
			_16 = function(b1, b2)
				_native.u8[0], _native.u8[1] = b1, b2
				return _native.u16[0]
			end,
			_32 = function(b1, b2, b3, b4)
				_native.u8[0], _native.u8[1], _native.u8[2], _native.u8[3] = b1, b2, b3, b4
				return _native.u32[0]
			end,
			_64 = function(b1, b2, b3, b4, b5, b6, b7, b8)
				_native.u8[0], _native.u8[1], _native.u8[2], _native.u8[3],
				_native.u8[4], _native.u8[5], _native.u8[6], _native.u8[7] = b1, b2, b3, b4, b5, b6, b7, b8
				return _native.u64
			end,
		}
	},

	be = {
		_out = {
			_16 = function(b1, b2) return b2, b1 end,
			_32 = function(w1, w2) return w2, w1 end,
			_64 = function(d1, d2) return d2, d1 end,
		},

		_in = {
			_16 = function(b1, b2)
				_native.u8[0], _native.u8[1] = b2, b1
				return _native.u16[0]
			end,
			_32 = function(b1, b2, b3, b4)
				_native.u8[0], _native.u8[1], _native.u8[2], _native.u8[3] = b4, b3, b2, b1
				return _native.u32[0]
			end,
			_64 = function(b1, b2, b3, b4, b5, b6, b7, b8)
				_native.u8[0], _native.u8[1], _native.u8[2], _native.u8[3],
				_native.u8[4], _native.u8[5], _native.u8[6], _native.u8[7] = b8, b7, b6, b5, b4, b3, b2, b1
				return _native.u64
			end,
		}
	}
}

_tags = {
	stop = 0,
	number = 1,
	string = 2,
	boolean = 3,
	table = 4,
}

_taggedWriters = {
	Blob.writeNumber,
	Blob.writeString,
	Blob.writeBool,
	Blob.writeTable,
}

_taggedReaders = {
	Blob.readNumber,
	Blob.readString,
	Blob.readBool,
	Blob.readTable,
}

return setmetatable({}, {
	__call = function(self, ...)
		local obj = setmetatable({}, { __index = Blob })
		obj:init(...)
		return obj
	end
})
